@@ -2,37 +2,38 @@ import copy
 import json
 import os
 import types
+from importlib.util import spec_from_loader
+from os import PathLike
 from os.path import join, splitext, expanduser
+from types import ModuleType
+from typing import Any, Dict, Iterator, Optional, Tuple, Type, Union
 
 from .env import Environment
 from .exceptions import UnknownFileType, UnpicklableConfigMember
 from .runners import Local
 from .terminals import WINDOWS
-from .util import debug, six, yaml
+from .util import debug, yaml
 
 
-if six.PY3:
-    try:
-        from importlib.machinery import SourceFileLoader
-    except ImportError:  # PyPy3
-        from importlib._bootstrap import _SourceFileLoader as SourceFileLoader
-
-    def load_source(name, path):
-        if not os.path.exists(path):
-            return {}
-        return vars(SourceFileLoader("mod", path).load_module())
+try:
+    from importlib.machinery import SourceFileLoader
+except ImportError:  # PyPy3
+    from importlib._bootstrap import (  # type: ignore[no-redef]
+        _SourceFileLoader as SourceFileLoader,
+    )
 
 
-else:
-    import imp
-
-    def load_source(name, path):
-        if not os.path.exists(path):
-            return {}
-        return vars(imp.load_source("mod", path))
+def load_source(name: str, path: str) -> Dict[str, Any]:
+    if not os.path.exists(path):
+        return {}
+    loader = SourceFileLoader("mod", path)
+    mod = ModuleType("mod")
+    mod.__spec__ = spec_from_loader("mod", loader)
+    loader.exec_module(mod)
+    return vars(mod)
 
 
-class DataProxy(object):
+class DataProxy:
     """
     Helper class implementing nested dict+attr access for `.Config`.
 
@@ -73,7 +74,12 @@ class DataProxy(object):
     )
 
     @classmethod
-    def from_data(cls, data, root=None, keypath=tuple()):
+    def from_data(
+        cls,
+        data: Dict[str, Any],
+        root: Optional["DataProxy"] = None,
+        keypath: Tuple[str, ...] = tuple(),
+    ) -> "DataProxy":
         """
         Alternate constructor for 'baby' DataProxies used as sub-dict values.
 
@@ -102,7 +108,7 @@ class DataProxy(object):
         obj._set(_keypath=keypath)
         return obj
 
-    def __getattr__(self, key):
+    def __getattr__(self, key: str) -> Any:
         # NOTE: due to default Python attribute-lookup semantics, "real"
         # attributes will always be yielded on attribute access and this method
         # is skipped. That behavior is good for us (it's more intuitive than
@@ -122,7 +128,7 @@ class DataProxy(object):
             err += "\n\nValid real attributes: {!r}".format(attrs)
             raise AttributeError(err)
 
-    def __setattr__(self, key, value):
+    def __setattr__(self, key: str, value: Any) -> None:
         # Turn attribute-sets into config updates anytime we don't have a real
         # attribute with the given name/key.
         has_real_attr = key in dir(self)
@@ -131,14 +137,14 @@ class DataProxy(object):
             # to our internal dict/cache
             self[key] = value
         else:
-            super(DataProxy, self).__setattr__(key, value)
+            super().__setattr__(key, value)
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[Dict[str, Any]]:
         # For some reason Python is ignoring our __hasattr__ when determining
         # whether we support __iter__. BOO
         return iter(self._config)
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
         # NOTE: Can't proxy __eq__ because the RHS will always be an obj of the
         # current class, not the proxied-to class, and that causes
         # NotImplemented.
@@ -149,24 +155,19 @@ class DataProxy(object):
         # itself just a dict.
         if isinstance(other, dict):
             other_val = other
-        return self._config == other_val
+        return bool(self._config == other_val)
 
-    # Make unhashable, because our entire raison d'etre is to be somewhat
-    # mutable. Subclasses with mutable attributes may override this.
-    # NOTE: this is mostly a concession to Python 2, v3 does it automatically.
-    __hash__ = None
-
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self._config)
 
-    def __setitem__(self, key, value):
+    def __setitem__(self, key: str, value: str) -> None:
         self._config[key] = value
         self._track_modification_of(key, value)
 
-    def __getitem__(self, key):
+    def __getitem__(self, key: str) -> Any:
         return self._get(key)
 
-    def _get(self, key):
+    def _get(self, key: str) -> Any:
         # Short-circuit if pickling/copying mechanisms are asking if we've got
         # __setstate__ etc; they'll ask this w/o calling our __init__ first, so
         # we'd be in a RecursionError-causing catch-22 otherwise.
@@ -186,7 +187,7 @@ class DataProxy(object):
             value = DataProxy.from_data(data=value, root=root, keypath=keypath)
         return value
 
-    def _set(self, *args, **kwargs):
+    def _set(self, *args: Any, **kwargs: Any) -> None:
         """
         Convenience workaround of default 'attrs are config keys' behavior.
 
@@ -200,24 +201,24 @@ class DataProxy(object):
         """
         if args:
             object.__setattr__(self, *args)
-        for key, value in six.iteritems(kwargs):
+        for key, value in kwargs.items():
             object.__setattr__(self, key, value)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "<{}: {}>".format(self.__class__.__name__, self._config)
 
-    def __contains__(self, key):
+    def __contains__(self, key: str) -> bool:
         return key in self._config
 
     @property
-    def _is_leaf(self):
+    def _is_leaf(self) -> bool:
         return hasattr(self, "_root")
 
     @property
-    def _is_root(self):
+    def _is_root(self) -> bool:
         return hasattr(self, "_modify")
 
-    def _track_removal_of(self, key):
+    def _track_removal_of(self, key: str) -> None:
         # Grab the root object responsible for tracking removals; either the
         # referenced root (if we're a leaf) or ourselves (if we're not).
         # (Intermediate nodes never have anything but __getitem__ called on
@@ -230,7 +231,7 @@ class DataProxy(object):
         if target is not None:
             target._remove(getattr(self, "_keypath", tuple()), key)
 
-    def _track_modification_of(self, key, value):
+    def _track_modification_of(self, key: str, value: str) -> None:
         target = None
         if self._is_leaf:
             target = self._root
@@ -239,11 +240,11 @@ class DataProxy(object):
         if target is not None:
             target._modify(getattr(self, "_keypath", tuple()), key, value)
 
-    def __delitem__(self, key):
+    def __delitem__(self, key: str) -> None:
         del self._config[key]
         self._track_removal_of(key)
 
-    def __delattr__(self, name):
+    def __delattr__(self, name: str) -> None:
         # Make sure we don't screw up true attribute deletion for the
         # situations that actually want it. (Uncommon, but not rare.)
         if name in self:
@@ -251,12 +252,12 @@ class DataProxy(object):
         else:
             object.__delattr__(self, name)
 
-    def clear(self):
+    def clear(self) -> None:
         keys = list(self.keys())
         for key in keys:
             del self[key]
 
-    def pop(self, *args):
+    def pop(self, *args: Any) -> Any:
         # Must test this up front before (possibly) mutating self._config
         key_existed = args and args[0] in self._config
         # We always have a _config (whether it's a real dict or a cache of
@@ -273,12 +274,12 @@ class DataProxy(object):
         # In all cases, return the popped value.
         return ret
 
-    def popitem(self):
+    def popitem(self) -> Any:
         ret = self._config.popitem()
         self._track_removal_of(ret[0])
         return ret
 
-    def setdefault(self, *args):
+    def setdefault(self, *args: Any) -> Any:
         # Must test up front whether the key existed beforehand
         key_existed = args and args[0] in self._config
         # Run locally
@@ -293,9 +294,9 @@ class DataProxy(object):
         self._track_modification_of(key, default)
         return ret
 
-    def update(self, *args, **kwargs):
+    def update(self, *args: Any, **kwargs: Any) -> None:
         if kwargs:
-            for key, value in six.iteritems(kwargs):
+            for key, value in kwargs.items():
                 self[key] = value
         elif args:
             # TODO: complain if arity>1
@@ -426,7 +427,7 @@ class Config(DataProxy):
     env_prefix = None
 
     @staticmethod
-    def global_defaults():
+    def global_defaults() -> Dict[str, Any]:
         """
         Return the core default settings for Invoke.
 
@@ -510,13 +511,13 @@ class Config(DataProxy):
 
     def __init__(
         self,
-        overrides=None,
-        defaults=None,
-        system_prefix=None,
-        user_prefix=None,
-        project_location=None,
-        runtime_path=None,
-        lazy=False,
+        overrides: Optional[Dict[str, Any]] = None,
+        defaults: Optional[Dict[str, Any]] = None,
+        system_prefix: Optional[str] = None,
+        user_prefix: Optional[str] = None,
+        project_location: Optional[PathLike] = None,
+        runtime_path: Optional[PathLike] = None,
+        lazy: bool = False,
     ):
         """
         Creates a new config object.
@@ -653,12 +654,12 @@ class Config(DataProxy):
         # a subroutine does so.
         self.merge()
 
-    def load_base_conf_files(self):
+    def load_base_conf_files(self) -> None:
         # Just a refactor of something done in unlazy init or in clone()
         self.load_system(merge=False)
         self.load_user(merge=False)
 
-    def load_defaults(self, data, merge=True):
+    def load_defaults(self, data: Dict[str, Any], merge: bool = True) -> None:
         """
         Set or replace the 'defaults' configuration level, from ``data``.
 
@@ -676,7 +677,7 @@ class Config(DataProxy):
         if merge:
             self.merge()
 
-    def load_overrides(self, data, merge=True):
+    def load_overrides(self, data: Dict[str, Any], merge: bool = True) -> None:
         """
         Set or replace the 'overrides' configuration level, from ``data``.
 
@@ -694,7 +695,7 @@ class Config(DataProxy):
         if merge:
             self.merge()
 
-    def load_system(self, merge=True):
+    def load_system(self, merge: bool = True) -> None:
         """
         Load a system-level config file, if possible.
 
@@ -711,7 +712,7 @@ class Config(DataProxy):
         """
         self._load_file(prefix="system", merge=merge)
 
-    def load_user(self, merge=True):
+    def load_user(self, merge: bool = True) -> None:
         """
         Load a user-level config file, if possible.
 
@@ -728,7 +729,7 @@ class Config(DataProxy):
         """
         self._load_file(prefix="user", merge=merge)
 
-    def load_project(self, merge=True):
+    def load_project(self, merge: bool = True) -> None:
         """
         Load a project-level config file, if possible.
 
@@ -750,7 +751,7 @@ class Config(DataProxy):
         """
         self._load_file(prefix="project", merge=merge)
 
-    def set_runtime_path(self, path):
+    def set_runtime_path(self, path: Optional[PathLike]) -> None:
         """
         Set the runtime config file path.
 
@@ -764,7 +765,7 @@ class Config(DataProxy):
         # if no loading has been attempted yet.)
         self._set(_runtime_found=None)
 
-    def load_runtime(self, merge=True):
+    def load_runtime(self, merge: bool = True) -> None:
         """
         Load a runtime-level config file, if one was specified.
 
@@ -782,7 +783,7 @@ class Config(DataProxy):
         """
         self._load_file(prefix="runtime", absolute=True, merge=merge)
 
-    def load_shell_env(self):
+    def load_shell_env(self) -> None:
         """
         Load values from the shell environment.
 
@@ -807,7 +808,9 @@ class Config(DataProxy):
         debug("Loaded shell environment, triggering final merge")
         self.merge()
 
-    def load_collection(self, data, merge=True):
+    def load_collection(
+        self, data: Dict[str, Any], merge: bool = True
+    ) -> None:
         """
         Update collection-driven config data.
 
@@ -822,7 +825,7 @@ class Config(DataProxy):
         if merge:
             self.merge()
 
-    def set_project_location(self, path):
+    def set_project_location(self, path: Union[PathLike, str, None]) -> None:
         """
         Set the directory path where a project-level config file may be found.
 
@@ -844,7 +847,9 @@ class Config(DataProxy):
         # Data loaded from the per-project config file.
         self._set(_project={})
 
-    def _load_file(self, prefix, absolute=False, merge=True):
+    def _load_file(
+        self, prefix: str, absolute: bool = False, merge: bool = True
+    ) -> None:
         # Setup
         found = "_{}_found".format(prefix)
         path = "_{}_path".format(prefix)
@@ -905,20 +910,19 @@ class Config(DataProxy):
         elif merge:
             self.merge()
 
-    def _load_yaml(self, path):
+    def _load_yaml(self, path: PathLike) -> Any:
         with open(path) as fd:
             return yaml.safe_load(fd)
 
-    def _load_yml(self, path):
-        return self._load_yaml(path)
+    _load_yml = _load_yaml
 
-    def _load_json(self, path):
+    def _load_json(self, path: PathLike) -> Any:
         with open(path) as fd:
             return json.load(fd)
 
-    def _load_py(self, path):
+    def _load_py(self, path: str) -> Dict[str, Any]:
         data = {}
-        for key, value in six.iteritems(load_source("mod", path)):
+        for key, value in (load_source("mod", path)).items():
             # Strip special members, as these are always going to be builtins
             # and other special things a user will not want in their config.
             if key.startswith("__"):
@@ -934,7 +938,7 @@ class Config(DataProxy):
             data[key] = value
         return data
 
-    def merge(self):
+    def merge(self) -> None:
         """
         Merge all config sources, in order.
 
@@ -959,7 +963,7 @@ class Config(DataProxy):
         debug("Deletions: {!r}".format(self._deletions))
         obliterate(self._config, self._deletions)
 
-    def _merge_file(self, name, desc):
+    def _merge_file(self, name: str, desc: str) -> None:
         # Setup
         desc += " config file"  # yup
         found = getattr(self, "_{}_found".format(name))
@@ -978,7 +982,7 @@ class Config(DataProxy):
             # the negative? Just a branch here based on 'name'?
             debug("{} not found, skipping".format(desc))
 
-    def clone(self, into=None):
+    def clone(self, into: Optional[Type["Config"]] = None) -> "Config":
         """
         Return a copy of this configuration object.
 
@@ -1010,16 +1014,8 @@ class Config(DataProxy):
         :returns:
             A `.Config`, or an instance of the class given to ``into``.
 
-        :raises:
-            ``TypeError``, if ``into`` is given a value and that value is not a
-            `.Config` subclass.
-
         .. versionadded:: 1.0
         """
-        # Sanity check for 'into'
-        if into is not None and not issubclass(into, self.__class__):
-            err = "'into' must be a subclass of {}!"
-            raise TypeError(err.format(self.__class__.__name__))
         # Construct new object
         klass = self.__class__ if into is None else into
         # Also allow arbitrary constructor kwargs, for subclasses where passing
@@ -1074,7 +1070,9 @@ class Config(DataProxy):
         new.merge()
         return new
 
-    def _clone_init_kwargs(self, into=None):
+    def _clone_init_kwargs(
+        self, into: Optional[Type["Config"]] = None
+    ) -> Dict[str, Any]:
         """
         Supply kwargs suitable for initializing a new clone of this object.
 
@@ -1101,7 +1099,7 @@ class Config(DataProxy):
             lazy=True,
         )
 
-    def _modify(self, keypath, key, value):
+    def _modify(self, keypath: Tuple[str, ...], key: str, value: str) -> None:
         """
         Update our user-modifications config level with new data.
 
@@ -1120,9 +1118,9 @@ class Config(DataProxy):
         excise(self._deletions, keypath + (key,))
         # Now we can add it to the modifications structure.
         data = self._modifications
-        keypath = list(keypath)
-        while keypath:
-            subkey = keypath.pop(0)
+        keypath_list = list(keypath)
+        while keypath_list:
+            subkey = keypath_list.pop(0)
             # TODO: could use defaultdict here, but...meh?
             if subkey not in data:
                 # TODO: generify this and the subsequent 3 lines...
@@ -1131,7 +1129,7 @@ class Config(DataProxy):
         data[key] = value
         self.merge()
 
-    def _remove(self, keypath, key):
+    def _remove(self, keypath: Tuple[str, ...], key: str) -> None:
         """
         Like `._modify`, but for removal.
         """
@@ -1140,9 +1138,9 @@ class Config(DataProxy):
         # inverse - remove from _deletions on modification.
         # TODO: may be sane to push this step up to callers?
         data = self._deletions
-        keypath = list(keypath)
-        while keypath:
-            subkey = keypath.pop(0)
+        keypath_list = list(keypath)
+        while keypath_list:
+            subkey = keypath_list.pop(0)
             if subkey in data:
                 data = data[subkey]
                 # If we encounter None, it means something higher up than our
@@ -1167,7 +1165,9 @@ class AmbiguousMergeError(ValueError):
     pass
 
 
-def merge_dicts(base, updates):
+def merge_dicts(
+    base: Dict[str, Any], updates: Dict[str, Any]
+) -> Dict[str, Any]:
     """
     Recursively merge dict ``updates`` into dict ``base`` (mutating ``base``.)
 
@@ -1224,19 +1224,19 @@ def merge_dicts(base, updates):
     return base
 
 
-def _merge_error(orig, new_):
+def _merge_error(orig: object, new: object) -> AmbiguousMergeError:
     return AmbiguousMergeError(
         "Can't cleanly merge {} with {}".format(
-            _format_mismatch(orig), _format_mismatch(new_)
+            _format_mismatch(orig), _format_mismatch(new)
         )
     )
 
 
-def _format_mismatch(x):
+def _format_mismatch(x: object) -> str:
     return "{} ({!r})".format(type(x), x)
 
 
-def copy_dict(source):
+def copy_dict(source: Dict[str, Any]) -> Dict[str, Any]:
     """
     Return a fresh copy of ``source`` with as little shared state as possible.
 
@@ -1248,17 +1248,17 @@ def copy_dict(source):
     return merge_dicts({}, source)
 
 
-def excise(dict_, keypath):
+def excise(dict_: Dict[str, Any], keypath: Tuple[str, ...]) -> None:
     """
     Remove key pointed at by ``keypath`` from nested dict ``dict_``, if exists.
 
     .. versionadded:: 1.0
     """
     data = dict_
-    keypath = list(keypath)
-    leaf_key = keypath.pop()
-    while keypath:
-        key = keypath.pop(0)
+    keypath_list = list(keypath)
+    leaf_key = keypath_list.pop()
+    while keypath_list:
+        key = keypath_list.pop(0)
         if key not in data:
             # Not there, nothing to excise
             return
@@ -1267,13 +1267,13 @@ def excise(dict_, keypath):
         del data[leaf_key]
 
 
-def obliterate(base, deletions):
+def obliterate(base: Dict[str, Any], deletions: Dict[str, Any]) -> None:
     """
     Remove all (nested) keys mentioned in ``deletions``, from ``base``.
 
     .. versionadded:: 1.0
     """
-    for key, value in six.iteritems(deletions):
+    for key, value in deletions.items():
         if isinstance(value, dict):
             # NOTE: not testing for whether base[key] exists; if something's
             # listed in a deletions structure, it must exist in some source
